/*
 *  plex86: run multiple x86 operating systems concurrently
 *  Copyright (C) 1999-2000  Kevin P. Lawton
 *
 *  io.c: IO oriented instructions.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */


#include "plex86.h"
#include "monitor.h"



  void
INSW_YvDX(vm_t *vm)
  /* input word/doubleword from port to string */
{
  Bit32u edi;
  unsigned int incr;

  if (vm->i.as_32)
    edi = G_EDI(vm);
  else
    edi = G_DI(vm);

  if (vm->i.os_32) {
    Bit32u value32=0;

    IOPermCheck(vm, G_DX(vm), 4);

    /* Write a zero to memory, to trigger any segment or page */
    /* faults before reading from IO port. */
    write_virtual_dword(vm, SRegES, edi, &value32);

    value32 = sysIOIn(vm, G_DX(vm), 4);

    /* no seg override allowed */
    write_virtual_dword(vm, SRegES, edi, &value32);
    incr = 4;
    }
  else {
    Bit16u value16=0;

    IOPermCheck(vm, G_DX(vm), 2);

    /* Write a zero to memory, to trigger any segment or page */
    /* faults before reading from IO port. */
    write_virtual_word(vm, SRegES, edi, &value16);

    value16 = sysIOIn(vm, G_DX(vm), 2);

    /* no seg override allowed */
    write_virtual_word(vm, SRegES, edi, &value16);
    incr = 2;
    }

  if (vm->i.as_32) {
    if (G_GetDF(vm))
      G_EDI(vm) -= incr;
    else
      G_EDI(vm) += incr;
    }
  else {
    if (G_GetDF(vm))
      G_DI(vm) -= incr;
    else
      G_DI(vm) += incr;
    }
}

  void
OUTSW_DXXv(vm_t *vm)
{
  unsigned seg;
  Bit32u esi;
  unsigned int incr;

  if (!NULL_SEG_REG(vm->i.seg)) {
    seg = vm->i.seg;
    }
  else {
    seg = SRegDS;
    }

  if (vm->i.as_32)
    esi = G_ESI(vm);
  else
    esi = G_SI(vm);

  if (vm->i.os_32) {
    Bit32u value32;

    IOPermCheck(vm, G_DX(vm), 4);

    read_virtual_dword(vm, seg, esi, &value32);

    sysIOOut(vm, G_DX(vm), 4, value32);
    incr = 4;
    }
  else {
    Bit16u value16;

    IOPermCheck(vm, G_DX(vm), 2);

    read_virtual_word(vm, seg, esi, &value16);

    sysIOOut(vm, G_DX(vm), 2, value16);
    incr = 2;
    }

  if (vm->i.as_32) {
    if (G_GetDF(vm))
      G_ESI(vm) -= incr;
    else
      G_ESI(vm) += incr;
    }
  else {
    if (G_GetDF(vm))
      G_SI(vm) -= incr;
    else
      G_SI(vm) += incr;
    }
}

  void
OUT_IbeAX(vm_t *vm)
{
  Bit32u port;
 
  port = vm->i.Ib;
 
  if (vm->i.os_32) {
    IOPermCheck(vm, port, 4);
    sysIOOut(vm, port, 4, G_EAX(vm));
    }
  else {
    IOPermCheck(vm, port, 2);
    sysIOOut(vm, port, 2, G_AX(vm));
    }
}

  void
IN_ALDX(vm_t *vm)
{
  unsigned port;

  port = G_DX(vm);
  IOPermCheck(vm, port, 1);
  G_AL(vm) = sysIOIn(vm, port, 1);
}

  void
IN_eAXDX(vm_t *vm)
{
  unsigned port;

  port = G_DX(vm);
  if (vm->i.os_32) {
    IOPermCheck(vm, port, 4);
    G_EAX(vm) = sysIOIn(vm, port, 4);
    }
  else {
    IOPermCheck(vm, port, 2);
    G_AX(vm) = sysIOIn(vm, port, 2);
    }
}
